import os
import traceback
from flask import Flask, request
from flask_restful import Api, Resource
from nproxypool.base.db import RedisPool
from nproxypool.utils.exceptions import PoolEmptyException
from nproxypool.utils.misc import get_sychronizer_config, get_redis_config


POOL = None


def get_second_level_pool(dir):
    synch_config = get_sychronizer_config(dir)
    redis_config = get_redis_config(dir)
    redis_config['db'] = synch_config['second_db']
    global POOL
    POOL = RedisPool(**redis_config)


class GetProxyRandomly(Resource):
    """
    'key' is a key in second-level pool 
    """
    @staticmethod
    def get():
        key = request.args.get('key')
        try:
            if key:
                proxy = POOL.z_random_get_one(key=key)
                res = {
                    'proxy': proxy
                }
            else:
                res = {
                    'errMsg': 'please input a key'
                }
        except PoolEmptyException:
            res = {
                'errMsg': 'proxy pool is empty: {}'.format(str(traceback.format_exc()))
            }
        except:
            res = {
                'errMsg': 'query error{}'.format(str(traceback.format_exc()))
            }
        return res


class GetProxyTotalNum(Resource):
    @staticmethod
    def get():
        _key = request.args.get('key')
        _min = int(request.args.get('min', '0'))
        _max = int(request.args.get('max', '100'))
        _min = 0 if (_min < 0 or _min > 100) else _min
        _max = 100 if (_max < _min or _max > 100) else _max
        res = {}
        if bytes(_key, encoding='utf-8') in POOL.z_get_all_keys():
            _total = POOL.z_get_total_num(_key, min_score=_min, max_score=_max)
            res[_key] = _total
        else:
            res['errMsg'] = 'invalid key: {}'.format(_key)
        return res


class GetProxyList(Resource):
    @staticmethod
    def get():
        _key = request.args.get('key')
        _min = int(request.args.get('min', '0'))
        _max = int(request.args.get('max', '100'))
        _min = 0 if (_min < 0 or _min > 100) else _min
        _max = 100 if (_max < _min or _max > 100) else _max
        res = {}
        if bytes(_key, encoding='utf-8') not in POOL.z_get_all_keys():
            res['errMsg'] = 'invalid key: {}'.format(_key)
        else:
            _list = POOL.z_get_total_num(_key, min_score=_min, max_score=_max, get_list=True)
            res[_key] = {
                'total': len(_list),
                'list': _list
            }
        return res


def make_app(dir):
    get_second_level_pool(dir)
    app = Flask(__name__)
    api = Api(app=app)
    api.add_resource(GetProxyRandomly, '/proxy/random/')
    api.add_resource(GetProxyTotalNum, '/proxy/total/')
    api.add_resource(GetProxyList, '/proxy/list/')
    return app



# ========================================================
# automatically remove proxies exist in second-level pools
# but not in the first-level one
# ========================================================
import time
import traceback
from nproxypool.base.db import RedisPool
from nproxypool.base.logger import get_logger


class Synchronizer(object):
    def __init__(self):
        self._first_level_pool = None
        self._second_level_pool = None
        self._second_level_pool_keys = []

    def init_sync(self, **kwargs):
        host = kwargs.get('host', 'localhost')
        port = kwargs.get('port', 6379)
        password = kwargs.get('password', '654321')
        first_db = kwargs.get('first_db', 0)
        second_db = kwargs.get('second_db', 1)
        second_db_keys = kwargs.get('second_db_keys', [])

        first_db_param = {
            'host': host,
            'port': port,
            'password': password,
            'db': first_db
        }
        second_db_param = {
            'host': host,
            'port': port,
            'password': password,
            'db': second_db
        }
        self._second_level_pool_keys = second_db_keys
        self._first_level_pool = RedisPool(**first_db_param)
        self._second_level_pool = RedisPool(**second_db_param)

    def synchronize(self, **kwargs):
        self.init_sync(**kwargs)
        first_level_pool_proxies = self._first_level_pool.get_all()
        for key in self._second_level_pool_keys:
            proxies_in_second_key = self._second_level_pool.z_get_all(key)
            for proxy in proxies_in_second_key:
                if proxy not in first_level_pool_proxies:
                    if isinstance(proxy, bytes):
                        proxy = str(proxy, encoding='utf-8')
                    print("!!!remove proxy: {}".format(proxy))
                    self._second_level_pool.z_delete(proxy, key)


def sync(**kwargs):
    to_sleep = kwargs.get('to_sleep', 300)
    file_dir = kwargs.get('log_file_path')
    logger = get_logger(
        name='synchronizer',
        filename='synchronizer.log',
        filedir=file_dir,
        level='warning',
        back_count=7
    )
    while True:
        try:
            s = Synchronizer()
            s.synchronize(**kwargs)
        except:
            logger.error(str(traceback.format_exc()))
        time.sleep(to_sleep)



import os
import logging
from logging import handlers, getLogger


class FileFilter(logging.Filter):
    def filter(self, record):
        try:
            filter_key = record.levelname
        except AttributeError:
            filter_key = None
        if filter_key in ['WARNING', 'ERROR', 'CRITICAL']:
            result = 1
        else:
            result = 0
        return result


class Logger(object):
    level_relations = {
        'notset': logging.NOTSET,
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'critical': logging.CRITICAL
    }

    def __init__(self, name='', filename=None, filedir=None, level='info', when='D', back_count=2, fmt=''):
        self.logger = getLogger(name=name)
        self.logger.setLevel(logging.INFO)
        if not fmt:
            fmt = '%(asctime)s - %(pathname)s[line:%(lineno)d] - %(levelname)s: %(message)s'

        format_str = logging.Formatter(fmt) 
        if filename and filedir:
            filename = os.path.join(filedir, filename)
            f = handlers.TimedRotatingFileHandler(
                filename=filename, when=when, interval=1, backupCount=back_count, encoding='utf-8')
            f.setFormatter(format_str)
            f.setLevel(self.level_relations.get(level))
            filter_ = FileFilter()
            f.addFilter(filter_)
            self.logger.addHandler(f)

        t = logging.StreamHandler()
        t.setLevel(logging.INFO)
        t.setFormatter(format_str)
        self.logger.addHandler(t)
        logging.getLogger('apscheduler').setLevel(logging.WARNING)

    def __call__(self):
        return self.logger


def get_logger(name='', filename=None, filedir=None, level='info', when='D', back_count=2, fmt=''):
    _logger = Logger(
        name=name,
        filename=filename,
        filedir=filedir,
        level=level,
        when=when,
        back_count=back_count,
        fmt=fmt
    )()
    return _logger



# =============================================
# validation verification of a proxy
# ==============================================
import time
import asyncio
import aiohttp
import traceback
from lxml import etree
from abc import abstractmethod
from user_agent import generate_user_agent
from nproxypool.base.logger import get_logger
from nproxypool.base.db import RedisPool


class Base(object):
    def __init__(self):
        self._name = 'Default'
        self._url = 'https://feed.baidu.com/feed/api/tab/gettabinfo'
        self._time_out = 5
        self._max_worker = 20
        self._status_code_allowed = [200, 302, 303]
        self._loop = asyncio.new_event_loop()
        self._failure_list = []
        self._success_list = []
        self._logger = None
        self._os_type = {
            True: "android",
            False: ("mac", "win")
        }

    async def requests(self, method, url=None, is_mobile=False, **kwargs):
        kwargs.setdefault('timeout', aiohttp.ClientTimeout(self._time_out))
        if "headers" in kwargs:
            kwargs["headers"].setdefault("user-agent", generate_user_agent(os=self._os_type[is_mobile]))
        else:
            kwargs["headers"] = {
                "user-agent": generate_user_agent(os=self._os_type[is_mobile])
            }
        async with aiohttp.ClientSession() as session:
            async with session.request(method, url, **kwargs) as r:
                resp = r
                content_type = r.content_type
                if "text" in content_type or "html" in content_type:
                    content = await r.text(encoding=r.charset)
                    resp.xpath = etree.HTML(content).xpath
                elif "json" in content_type:
                    content = await r.json(encoding=r.charset)
                else:
                    content = await r.read()
                if r.status not in self._status_code_allowed:
                    self._logger.info('STATUS_CODE_NOT_ALLOWED ERROR')
                    raise Exception('!STATUS CODE ERROR!')
        return content, resp

    async def verify_worker(self, worker_name, queue):
        while True:
            proxy = await queue.get()
            _proxy = 'http://{}'.format(proxy) if not proxy.startswith('http') else proxy
            try:
                _start = time.monotonic()
                await self.requests('GET', self._url, proxy=_proxy)
                _time_consumed = round(time.monotonic() - _start, 2)
                self._logger.info(f'>>>{worker_name}-ip: {_proxy}, time_consumed: {_time_consumed}')
                self._success_list.append(proxy)
            except:
                self._failure_list.append(proxy)
                self._logger.info(f'!!!{worker_name}-ip: {_proxy}, FAILED')
            queue.task_done()

    async def verify(self, proxy_list):
        queue = asyncio.Queue()

        for _proxy in proxy_list:
            proxy = _proxy.decode()
            queue.put_nowait(proxy)

        tasks = []
        for i in range(self._max_worker):
            task = self._loop.create_task(self.verify_worker(f'{self._name}-{i}', queue))
            tasks.append(task)

        await queue.join()
        for task in tasks:
            task.cancel()
        await asyncio.gather(*tasks, return_exceptions=True)

    @abstractmethod
    def init_verifier(self, **kwargs):
        pass

    @abstractmethod
    def run(self, **kwargs):
        pass


class Verifier(Base):
    def __init__(self):
        super(Verifier, self).__init__()
        self._redis_host = ''
        self._redis_port = ''
        self._password = ''
        self._from_db = 0
        self._from_db_key = ''
        self._is_from_db_zset = False
        self._to_db = 1
        self._to_db_key = ''
        self._is_to_db_zset = False
        self._score_threshold = 30
        self._score_decrease_by = 30
        self._redis_uri = "redis://:{password}@{host}:{port}/{db}"

    def init_verifier(self, **kwargs):
        """
        Initiate a verifier
        """
        self._redis_host = kwargs['host']
        self._redis_port = kwargs['port']
        self._password = kwargs['password']
        self._name = kwargs.get('name', 'Default')
        self._url = kwargs.get('url', 'https://feed.baidu.com/feed/api/tab/gettabinfo')
        self._max_worker = kwargs.get('max_worker', 20)
        self._time_out = kwargs.get('time_out', 5)
        self._from_db = kwargs.get('from_db', 0)
        self._from_db_key = kwargs.get('from_db_key', '')
        self._is_from_db_zset = kwargs.get('is_from_db_zset', False)
        self._to_db = kwargs.get('to_db', 1)
        self._to_db_key = kwargs.get('to_db_key', '')
        self._is_to_db_zset = kwargs.get('is_to_db_zset', False)
        self._score_threshold = kwargs.get('score_threshold', 30)
        self._score_decrease_by = kwargs.get('score_decrease_by', 30)

    def _delete_or_decrease(self, proxy: str, client: RedisPool, zset_key: str = ''):
        """
        Delete a proxy or decrease its score.
        """

        if not zset_key:
            value = client.get_one(proxy)
            if value:
                _, score = value.split(';')
                new_score = int(score) - self._score_decrease_by
                if new_score <= self._score_threshold:
                    client.delete_one(proxy)
                else:
                    client.update_one(
                        proxy,
                        new_score,
                        old_value=value
                    )
            else:
                client.z_decrease(
                    proxy,
                    zset_key,
                    score_threshold=self._score_threshold,
                    to_decr=self._score_decrease_by
                )

    def _insert_or_increase(self, proxy: str, client: RedisPool, zset_key: str = ''):
        """
        Insert a proxy or increase its score.
        """
        if not zset_key:
            value = client.get_one(proxy)
            if value:
                client.update_one(
                    proxy,
                    score=100,
                    old_value=value
                )
            else:
                client.insert_one(
                    proxy,
                    score=60
                )
        else:
            client.z_increase(
                proxy,
                zset_key
            )

    def _update(self, client: RedisPool, zset_key: str = ''):
        for proxy in self._failure_list:
            self._delete_or_decrease(
                proxy,
                client,
                zset_key=zset_key
            )
        for proxy in self._success_list:
            self._insert_or_increase(
                proxy,
                client,
                zset_key=zset_key
            )

    def _key_to_key_verify(self):
        """
        first-level pool to first-level pool update.
        """
        params = {
            'host': self._redis_host,
            'port': self._redis_port,
            'db': self._from_db,
            'password': self._password
        }
        client = RedisPool(**params)
        proxy_list = client.get_all()
        if proxy_list:
            self._loop.run_until_complete(self.verify(proxy_list))
            self._update(client)

    def _key_to_zset_verify(self):
        """
        first-level pool to second-level pool update.
        """
        params1 = {
            'host': self._redis_host,
            'port': self._redis_port,
            'db': self._from_db,
            'password': self._password
        }
        params2 = {
            'host': self._redis_host,
            'port': self._redis_port,
            'db': self._to_db,
            'password': self._password
        }
        client1 = RedisPool(**params1)
        client2 = RedisPool(**params2)
        proxy_list = client1.get_all()
        if proxy_list:
            self._loop.run_until_complete(self.verify(proxy_list))
            self._update(client2, zset_key=self._to_db_key)

    def _zest_to_zset_verify(self):
        """
        second-level pool to second-level pool update.
        """
        params = {
            'host': self._redis_host,
            'port': self._redis_port,
            'db': self._from_db,
            'password': self._password
        }
        client = RedisPool(**params)
        proxy_list = client.z_get_all(self._from_db_key)
        if proxy_list:
            self._loop.run_until_complete(self.verify(proxy_list))
            self._update(client, zset_key=self._to_db_key)

    def run(self, **kwargs):
        self.init_verifier(**kwargs)
        self._failure_list = []
        self._success_list = []
        if self._from_db == self._to_db and not self._is_from_db_zset:
            self._key_to_key_verify()
        elif self._is_to_db_zset and not self._is_from_db_zset:
            self._key_to_zset_verify()
        else:
            self._zest_to_zset_verify()


def run_verifier(**kwargs):
    time_sleep = kwargs.get('to_sleep', 10)
    _name = kwargs.get('name', 'Default')
    _name = 'verifier_{}'.format(_name)
    file_name = '{}.log'.format(_name)
    file_dir = kwargs.get('log_file_path')
    v_logger = get_logger(
        name=_name,
        filename=file_name,
        filedir=file_dir,
        level='warning',
        back_count=7
    )
    while True:
        try:
            v = Verifier()
            v._logger = v_logger
            v.run(**kwargs)
        except:
            v_logger.error(str(traceback.format_exc()))
        time.sleep(time_sleep)



import os
import sys
from nproxypool.utils.run import run_pool, run_server, run_generate


def show_usage():
    print('usage:')
    print('mypool gen:        to generate a new proxypool project.')
    print('mypool run_pool:   to run the proxypool.')
    print('mypool run_server: to run server for the proxypool.')
    sys.exit(1)


def execute():
    args = sys.argv[1:]
    if not args:
        show_usage()
    dir = os.path.abspath('.')
    cmdname = args[0]
    if cmdname == 'gen':
        if len(args) > 1:
            name = args[1]
        else:
            name = 'proxypool'
        run_generate(dir, name=name)
    elif cmdname == 'run_pool':
        run_pool(dir)
    elif cmdname == 'run_server':
        run_server(dir)
    else:
        show_usage()



from abc import abstractmethod
from nproxypool.base.db import RedisPool


class ProxyFetcherBase(object):
    def __init__(self, **kwargs):
        self._client = RedisPool(**kwargs)

    @abstractmethod
    def run(self):
        pass



# Redis Client to Save IPs
# -----------------------------------------
# every IP has an initial score as 60, decreases by 30 when a verification failure occurred,
# increase to 100 when a verification success occurred. An IP is removed when its score reach below 20.
#
# The score of an IP is decreased by 1 when fetched by a client.
# One ip is chosen among the highest scored ones when fetched by a client.
# -----------------------------------------
import time
from random import choice
from redis import ConnectionPool, StrictRedis
from nproxypool.utils.exceptions import PoolEmptyException


class RedisPoolBase(object):
    def __init__(self, **kwargs):
        self._kwargs = kwargs
        self._redis_uri = "redis://:{password}@{host}:{port}/{db}"
        self._client = None
        self._get_client()
        self._value_pattern = '{};{}'

    def __del__(self):
        if self._client:
            self._client.connection_pool.disconnect()

    def _get_client(self):
        if not self._kwargs.get('uri'):
            password = self._kwargs.get('password', '')
            host = self._kwargs.get('host', 'localhost')
            port = self._kwargs.get('port', '6379')
            db = self._kwargs.get('db', 1)
            redis_uri = self._redis_uri.format(
                password=password,
                host=host,
                port=port,
                db=db
            )
        else:
            redis_uri = self._kwargs['uri']
        pool = ConnectionPool.from_url(redis_uri)
        self._client = StrictRedis(connection_pool=pool)

    @staticmethod
    def _bytes_to_str(_value):
        res = _value
        if isinstance(_value, bytes):
            res = str(_value, encoding="utf-8")
        return res

    # ====================================
    #  first-level pool operations
    # ====================================
    def get_all(self):
        return self._client.keys()

    def get_one(self, key):
        """
        get one proxy from first-level pool
        key: proxy
        """
        r = self._client.get(key)
        return r if not r else self._bytes_to_str(r)

    def insert_one(self, key, score=60, expire=600):
        """
        insert one proxy to first-level pool
        key: proxy
        score: initial score
        expire: expire time
        """
        if not self.get_one(key):
            expire_stamp = int(time.time() + expire)
            value = self._value_pattern.format(expire_stamp, score)
            self._client.set(key, value)
            self._client.expireat(key, expire_stamp)

    def update_one(self, key, score=60, old_value=''):
        """
        update score but never change the expire time.
        """
        old_value = old_value or self.get_one(key)
        _expire = old_value.split(';')[0]
        expire_stamp = int(_expire) if _expire else int(time.time() - 300)
        new_value = self._value_pattern.format(expire_stamp, score)
        self._client.set(key, new_value)
        self._client.expireat(key, expire_stamp)

    def delete_one(self, key):
        self._client.delete(key)


class RedisPool(RedisPoolBase):
    def __init__(self, **kwargs):
        super(RedisPool, self).__init__(**kwargs)

    # ====================================
    #  second-level pool(zset) operations
    # ====================================
    def z_get_all_keys(self):
        return self._client.keys()

    def z_get_all(self, key):
        return self._client.zrangebyscore(key, -60, 100)

    def z_get(self, proxy, key):
        """
        query score or test whether the proxy is existed in a zset
        """
        return self._client.zscore(key, proxy)

    def z_add(self, proxy, key, score):
        if not self.z_get(proxy, key):
            _map = {proxy: score}
            self._client.zadd(key, _map)

    def z_increase(self, proxy, key):
        _map = {proxy: 100}
        self._client.zadd(key, _map)

    def z_delete(self, proxy, key):
        self._client.zrem(key, proxy)

    def z_decrease(self, proxy, key, score_threshold=20, to_decr=30):
        """
        score_threshold: threshold for proxy delete
        to_decr: score to decrease by when a verification failure occurred
        """
        score = self.z_get(proxy, key)
        if score:
            new_score = score - to_decr
            if new_score <= score_threshold:
                self.z_delete(proxy, key)
            else:
                self._client.zincrby(
                    name=key,
                    value=proxy,
                    amount=0 - to_decr
                )

    def z_random_get_one(self, key='baidu'):
        res = self._client.zrangebyscore(key, 100, 100)
        if not res:
            res = self._client.zrangebyscore(key, 85, 100)
        if res:
            proxy = choice(res)
            proxy = self._bytes_to_str(proxy)
            self.z_decrease(proxy, key, to_decr=1)
            return proxy
        else:
            raise PoolEmptyException()

    def z_get_total_num(self, key, min_score=0, max_score=100, get_list=False):
        res = self._client.zrangebyscore(key, min_score, max_score)
        if not get_list:
            return len(res)
        else:
            return [self._bytes_to_str(i) for i in res]



# ===========================
# run pool or server
# ===========================
import nproxypool
import os
import sys
import time
import signal
import traceback
import importlib
from copy import deepcopy
from multiprocessing import Process
from werkzeug.debug import DebuggedApplication
from nproxypool.base.verifier import run_verifier
from nproxypool.base.synchronizer import sync
from nproxypool.base.proxy_fetcher import ProxyFetcherBase
from nproxypool.utils.misc import get_redis_config, get_verifier_config
from nproxypool.utils.misc import get_sychronizer_config, get_fetcher_config
from nproxypool.utils.misc import get_logfile_dir, get_gunicorn_config
from nproxypool.utils.misc import GunicornStandalone
from nproxypool.base.server import make_app
from nproxypool.base.logger import get_logger
from nproxypool.utils.misc import copytree


# close subprocess when parent process is closed
def term(sig_num, addtion):
    print('term current pid is %s, group id is %s' % (os.getpid(), os.getpgrp()))
    os.killpg(os.getpgid(os.getpid()), signal.SIGKILL)


signal.signal(signal.SIGTERM, term)
print('master pid is %s' % os.getpid())


def run_fetcher(**kwargs):
    file_dir = kwargs.get('log_file_path')
    logger = get_logger(
        name='fetcher',
        filename='fetcher.log',
        filedir=file_dir,
        level='warning',
        back_count=7
    )
    to_sleep = kwargs.get('to_sleep', 30)
    fetcher_module = importlib.import_module('my_fetcher')
    for attr in dir(fetcher_module):
        v = getattr(fetcher_module, attr)
        if type(v).__name__ == 'type' and issubclass(v, ProxyFetcherBase) and attr != 'ProxyFetcherBase':
            try:
                fetcher = v(**kwargs)
                fetcher.run()
            except:
                logger.error(str(traceback.format_exc()))
    time.sleep(to_sleep)


def run_pool(dir):
    """
    dir: project dir
    """
    if dir not in sys.path:
        sys.path.append(dir)
    redis_params = get_redis_config(dir)
    log_file_dir = get_logfile_dir(dir)

    # fetcher process
    kwargs_fetcher = get_fetcher_config(dir)
    kwargs_fetcher.update(redis_params)
    kwargs_fetcher['log_file_path'] = log_file_dir
    p = Process(target=run_fetcher, kwargs=kwargs_fetcher)
    p.daemon = True
    p.start()

    # verifier processes
    verifier_conf_list = get_verifier_config(dir)
    for verifier_conf in verifier_conf_list:
        kwargs_v = deepcopy(verifier_conf)
        kwargs_v.update(redis_params)
        kwargs_v['log_file_path'] = log_file_dir
        t = Process(target=run_verifier, kwargs=kwargs_v)
        t.daemon = True
        t.start()

    # synchronizer process
    kwargs_sync = get_sychronizer_config(dir)
    kwargs_sync['log_file_path'] = log_file_dir
    p1 = Process(target=sync, kwargs=kwargs_sync)
    p1.daemon = True
    p1.start()

    while True:
        time.sleep(1)


def run_server(dir):
    options = get_gunicorn_config(dir)
    app = make_app(dir)
    if options.get('debug'):
        options["workers"] = 1
        app.wsgi_app = DebuggedApplication(app.wsgi_app, True)
        print(' * Launching in DEBUG mode')
        print(' * Serving Flask using a single worker "{}"'.format(app.import_name))
        options["reload"] = bool(options['debug'])
    else:
        print(' * Launching in Production Mode')
        print(' * Serving Flask with {} worker(s) "{}"'.format(
            options["workers"], app.import_name
        ))
    server = GunicornStandalone(app, options=options)
    server.run()


def run_generate(dir, name='proxypool'):
    tempfile_path = os.path.join(
        os.path.dirname(nproxypool.__file__),
        'templates/pool'
    )
    copytree(tempfile_path, os.path.join(dir, name))



import requests
from scrapy import Selector
from urllib.parse import urljoin
from user_agent import generate_user_agent
from nproxypool.base.proxy_fetcher import ProxyFetcherBase


class ProxyFetcherIhuan(ProxyFetcherBase):
    def __init__(self, **kwargs):
        super(ProxyFetcherIhuan, self).__init__(**kwargs)
        self._url = 'https://ip.ihuan.me/'
        self._user_agent = generate_user_agent(os='win')

    def run(self):
        headers = {
            'user-agent': self._user_agent
        }
        r = requests.get(self._url, headers=headers)
        s = Selector(text=r.text)
        page_list = s.xpath('//ul[@class="pagination"]//a/@href').extract()
        for _url in page_list:
            url = urljoin(self._url, _url)
            r = requests.get(url, headers=headers)
            s = Selector(text=r.text)
            labels = s.xpath('//div[@class="table-responsive"]//tr')
            for tr in labels:
                ip = tr.xpath('./td[1]/a/text()').get()
                port = tr.xpath('./td[2]/text()').get()
                if ip and port:
                    proxy = f'{ip}:{port}'
                    self._client.insert_one(proxy)


if __name__ == '__main__':
    p = ProxyFetcherIhuan()
    p.run()



class PoolEmptyException(Exception):

    def __init__(self):
        Exception.__init__(self)

    def __str__(self):
        return repr('Proxy pool is empty.')


class NotInProject(Exception):

    def __init__(self):
        Exception.__init__(self)

    def __str__(self):
        return repr('Please execute this command in a proxypool project.')



from setuptools import setup, find_packages


install_requires = [
    'Scrapy>=2.4.1',
    'Flask>=1.1.2',
    'Flask-RESTful>=0.3.8',
    'user-agent>=0.1.9',
    'gunicorn>=20.0.4',
    'redis>=3.5.3',
    'aiohttp>=3.7.3',
    'Werkzeug>=1.0.1'
]


setup(
    name='Nproxypool',
    version='1.0.2',
    url='https://github.com/moqsien/nproxypool',
    project_urls={
        'Documentation': 'https://github.com/moqsien/nproxypool/blob/main/README.md',
        'Source': 'https://github.com/moqsien/nproxypool',
        'Tracker': 'https://github.com/moqsien/nproxypool',
    },
    description='A Proxypool',
    author='MoQsien',
    maintainer='MoQsien',
    maintainer_email='moqsien@foxmail.com',
    packages=find_packages(exclude=('tests', 'tests.*', 'example')),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': ['npool = nproxypool.cmdline:execute']
    },
    python_requires='>=3.6',
    install_requires=install_requires,
    extras_require={}
)



import os
from configparser import ConfigParser
from stat import S_IWUSR as OWNER_WRITE_PERMISSION
from shutil import ignore_patterns, copy2, copystat
import gunicorn.app.base
from nproxypool.utils.exceptions import NotInProject


IGNORE = ignore_patterns('*.pyc', '__pycache__', '.svn')


def _make_writable(path):
    current_permissions = os.stat(path).st_mode
    os.chmod(path, current_permissions | OWNER_WRITE_PERMISSION)


class GunicornStandalone(gunicorn.app.base.BaseApplication):
    def __init__(self, application, options=None):
        """ Construct the Application. Default gUnicorn configuration is loaded """

        self.application = application
        self.options = options or {}
        print(self.options)

        # if port, or host isn't set-- run from os.environments
        #
        super(GunicornStandalone, self).__init__()

    def init(self, parser, opts, args):
        """ Apply our custom settings """

        cfg = {}
        for k, v in self.options.items():
            if k.lower() in self.cfg.settings and v is not None:
                cfg[k.lower()] = v
        return cfg

    def load_config(self):
        config = dict([(key, value) for key, value in self.options.items()
                       if key in self.cfg.settings and value is not None])
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application


def get_config_object(dir):
    file_path = os.path.join(dir, './pool.cfg')
    if not os.path.exists(file_path):
        raise NotInProject()
    config = ConfigParser()
    config.read(file_path)
    return config


def get_gunicorn_config(dir):
    config = get_config_object(dir)
    g_conf = dict()
    section = 'gunicorn'
    for opt in config.options(section):
        if opt in ('pidfile', 'accesslog', 'errorlog'):
            g_conf[opt] = os.path.join(dir, config.get(section, opt))
        else:
            g_conf[opt] = config.get(section, opt)
    return g_conf


def get_redis_config(dir):
    config = get_config_object(dir)
    conf_redis = dict()
    conf_redis['host'] = config.get('redis', 'host')
    conf_redis['port'] = config.getint('redis', 'port')
    conf_redis['password'] = config.get('redis', 'password')
    return conf_redis


def get_fetcher_insert_db(dir):
    config = get_config_object(dir)
    for section in config.sections():
        if 'init' in section:
            return dict(db=config.getint(section, 'from_db'))


def get_logfile_dir(dir):
    config = get_config_object(dir)
    logfile_dir_name = config.get('logfile', 'dirname')
    if not logfile_dir_name.startswith('./'):
        logfile_dir_name = './{}'.format(logfile_dir_name)
    return os.path.join(dir, logfile_dir_name)


def get_verifier_config(dir):
    config = get_config_object(dir)
    conf_list = []
    opt_bool = (
        'is_from_db_zset',
        'is_to_db_zset'
    )
    opt_int = (
        'from_db',
        'to_db',
        'score_threshold',
        'score_decrease_by',
        'time_out',
        'to_sleep',
        'max_worker'
    )
    for section in config.sections():
        s = dict()
        if section in ('redis',):
            continue
        for opt in config.options(section):
            if opt in opt_bool:
                s[opt] = config.getboolean(section, opt)
            elif opt in opt_int:
                s[opt] = config.getint(section, opt)
            else:
                s[opt] = config.get(section, opt)
        conf_list.append(s)
    return conf_list


def get_sychronizer_config(dir):
    config = get_config_object(dir)
    conf_sync = dict()
    second_db_keys = []
    for section in config.sections():
        if section in ('redis',):
            for opt in config.options(section):
                conf_sync[opt] = config.get(section, opt)
        elif 'init' in section:
            conf_sync['first_db'] = config.getint(section, 'from_db')
        elif 'sync' in section:
            conf_sync['to_sleep'] = config.getint(section, 'to_sleep')
        elif 'verifier' in section:
            conf_sync['second_db'] = config.getint(section, 'to_db')
            second_db_keys.append(config.get(section, 'to_db_key'))
        else:
            pass
    conf_sync['second_db_keys'] = list(set(second_db_keys))
    return conf_sync


def get_fetcher_config(dir):
    config = get_config_object(dir)
    conf_fetcher = dict()
    conf_fetcher['to_sleep'] = config.getint('fetcher', 'to_sleep')
    for section in config.sections():
        if 'init' in section:
            conf_fetcher['db'] = config.getint(section, 'from_db')
    return conf_fetcher


def copytree(src, dst):
    ignore = IGNORE
    names = os.listdir(src)
    ignored_names = ignore(src, names)
    if not os.path.exists(dst):
        os.makedirs(dst)
    for name in names:
        if name in ignored_names:
            continue
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        if os.path.isdir(srcname):
            copytree(srcname, dstname)
        else:
            copy2(srcname, dstname)
            _make_writable(dstname)
    copystat(src, dst)
    _make_writable(dst)
